
// source: wikipedia article on Kahan sum

#include <stdio.h>
#include <fenv.h>
//#pragma STDC FENV_ACCESS ON

// classical summation of array items
float sumf(float a[], unsigned size) {
    float sum = 0.0;
    for (int i = 0; i < size; ++i) {
        sum += a[i];
    }
    return sum;
}

double sum(double a[], unsigned size) {
    double sum = 0.0;
    for (int i = 0; i < size; ++i) {
        sum += a[i];
    }
    return sum;
}

long double suml(long double a[], unsigned size) {
    long double sum = 0.0;
    for (int i = 0; i < size; ++i) {
        sum += a[i];
    }
    return sum;
}

// Kahan sum implementations
float kahan_sumf(float a[], unsigned size) {
    float sum = 0.0;
    float c = 0.0;              // A running compensation for lost low-order bits.
    for (int i = 0; i < size; ++i) {
        float y = a[i] - c;     // So far, so good: c is zero.
        float t = sum + y;      // Alas, sum is big, y small, so low-order digits of y are lost.
        c = (t - sum) - y;      // (t - sum) cancels the high-order part of y; subtracting y recovers negative (low part of y)
        sum = t;                // Algebraically, c should always be zero.  Beware overly-aggressive optimizing compilers!
    }                           // Next time around, the lost low part will be added to y in a fresh attempt.
    return sum;
}

double kahan_sum(double a[], unsigned size) {
    double sum = 0.0;
    double c = 0.0;             // A running compensation for lost low-order bits.
    for (int i = 0; i < size; ++i) {
        double y = a[i] - c;    // So far, so good: c is zero.
        double t = sum + y;     // Alas, sum is big, y small, so low-order digits of y are lost.
        c = (t - sum) - y;      // (t - sum) cancels the high-order part of y; subtracting y recovers negative (low part of y)
        sum = t;                // Algebraically, c should always be zero.  Beware overly-aggressive optimizing compilers!
    }                           // Next time around, the lost low part will be added to y in a fresh attempt.
    return sum;
}

long double kahan_suml(long double a[], unsigned size) {
    long double sum = 0.0;
    long double c = 0.0;
    for (int i = 0; i < size; ++i) {
        long double y = a[i] - c;
        long double t = sum + y;
        c = (t - sum) - y;
        sum = t;
    }
    return sum;
}

// Testing data:
#define N 10
float af[N] = { 0.1F, 0.1F, 0.1F, 0.1F, 0.1F, 0.1F, 0.1F, 0.1F, 0.1F, 0.1F, };
double ad[N] = { 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, };
long double al[N] = { 0.1L, 0.1L, 0.1L, 0.1L, 0.1L, 0.1L, 0.1L, 0.1L, 0.1L, 0.1L, };
// test wrong literal type initialization:
// long double al[N] = { 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, };

// Run basic tests:
int main(void) {
    printf("Kahan sum vs classic:\n");
#ifdef TEST1
    fesetround(FE_TOWARDZERO);
#endif
    printf("float:\n");
    float fsum = kahan_sumf(af, N);
    printf("%.10f\t", fsum);
    printf("delta=%g\n", fsum - 1.0F);

    fsum = sumf(af, N);
    printf("%.10f\t", fsum);
    printf("delta=%g\n", fsum - 1.0F);

    printf("double:\n");
    double dsum = kahan_sum(ad, N);
    printf("%.20f\t", dsum);
    printf("delta=%g\n", dsum - 1.0);

    dsum = sum(ad, N);
    printf("%.20f\t", dsum);
    printf("delta=%g\n", dsum - 1.0);

    printf("long double:\n");
    long double l1 = kahan_suml(al, N);
    printf("%.25Lf\t", l1);
    printf("delta=%Lg\n", l1 - 1.0L);

    l1 = suml(al, N);
    printf("%.25Lf\t", l1);
    printf("delta=%Lg\n", l1 - 1.0L);
}
